package de.desy.acop.video.analysis.interferometer;

import java.awt.image.BufferedImage;
import java.util.Arrays;

import org.apache.commons.math.analysis.MultivariateRealFunction;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.acop.video.analysis.AnalysisType;
import de.desy.acop.video.analysis.FittingUtilities;
import de.desy.acop.video.analysis.FittingUtilities.InterferometerCostFunctionBessel;
import de.desy.acop.video.analysis.FittingUtilities.InterferometerCostFunctionTrigonometric;
import de.desy.acop.video.analysis.ImageAnalyzer;
import de.desy.acop.video.timageio.TBufferedImage;
import de.desy.tine.types.IMAGE;

/**
 * <code>InterferometerImageAnalyzer</code> fits the interferometer function to the side views.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class InterferometerImageAnalyzer extends ImageAnalyzer {

    public static enum Type {
        //for backward compatibility names are what they are
        HORIZONTAL, VERTICAL, BOTH, BESSEL, BESSEL_HORIZONTAL, BESSEL_VERTICAL
    }

    private static double[] NAN_ARRAY = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN,
            Double.NaN};
    private double[] previousStartValuesIf = null;
    private double[] previousStartValuesIf2 = null;
    private static final double DELTA = 0.05;
    private static final double COST_MIN = 3;
    private double currentCostMin = COST_MIN;
    private final Type interferometerType;
    private final boolean useBesselFunction;
    // Image stacking
    private int imagesInRotation = 0;
    private int numOfImages;
    private ImageDataWrapper[] images;
    private double[] newImageValues;
    private double[] rawImageValues;
    private int[] newImageData;
    private BufferedImage newBufferedImage;
    private int lastImageIndex = 0;
    private int oldImageWidth = 0;
    private int oldImageHeight = 0;
    // number of pixels from the center to each direction that are taken into account when summing the data for the fit
    private static final int HALFROIWIDTH = 5;
    // use sideview to find max or use absolute max
    private static final boolean USE_SIDE_VIEW_MAX = true;

    /**
     * Image wrapper class which contains image data, size and center coordinates and can perform centering to
     * coordinates.
     *
     * @author nsmrkolj
     */
    private class ImageDataWrapper {

        double[] values;
        int width;
        int height;
        int centerX;
        int centerY;

        /**
         * Create new image wrapper
         *
         * @param values Image data
         * @param width of width image
         * @param height of height image
         */
        ImageDataWrapper(double[] values, int width, int height) {
            this.values = values;
            this.width = width;
            this.height = height;
            calculateImageCenter();
        }

        private void calculateImageCenter() {
            double maxHorisontalValue = -Double.MAX_VALUE;
            double maxVerticalValue = -Double.MAX_VALUE;
            int maxHorisontalIndex = 0;
            int maxVerticalIndex = 0;
            if (USE_SIDE_VIEW_MAX) {
                double[] horisontalAxis = new double[width];
                double[] verticalAxis = new double[height];
                for (int i = values.length - 1; i > -1; --i) {
                    horisontalAxis[i % width] += values[i];
                    verticalAxis[i / width] += values[i];
                }
                for (int i = width - 1; i > -1; --i) {
                    if (horisontalAxis[i] > maxHorisontalValue) {
                        maxHorisontalValue = horisontalAxis[i];
                        maxHorisontalIndex = i;
                    }
                }
                for (int i = height - 1; i > -1; --i) {
                    if (verticalAxis[i] > maxVerticalValue) {
                        maxVerticalValue = verticalAxis[i];
                        maxVerticalIndex = i;
                    }
                }
            } else {
                for (int i = values.length - 1; i > -1; --i) {
                    if (values[i] > maxHorisontalValue) {
                        maxHorisontalValue = values[i];
                        maxHorisontalIndex = i % width;
                        maxVerticalIndex = i / width;
                    }
                }
            }
            centerX = maxHorisontalIndex;
            centerY = maxVerticalIndex;
        }

        /**
         * Stack image stored in ImageDataWrapper to passed image.
         *
         * @param x coordinate of center
         * @param y coordinate of center
         * @param newImageValues image to stack with
         */
        void stackImages(int x, int y, double[] newImageValues, double[] rawImage) {
            int offsetX = x - centerX;
            int offsetY = y - centerY;
            for (int i = 0; i < height; i++) {
                for (int j = 0; j < width; j++) {
                    int xCoord = 0;
                    int yCoord = 0;
                    if (i - offsetY < 0) {
                        xCoord = 0;
                    } else if (i - offsetY >= height) {
                        xCoord = height - 1;
                    } else {
                        xCoord = i - offsetY;
                    }
                    if (j - offsetX < 0) {
                        yCoord = 0;
                    } else if (j - offsetX >= width) {
                        yCoord = width - 1;
                    } else {
                        yCoord = j - offsetX;
                    }
                    newImageValues[i * width + j] += values[xCoord * width + yCoord];
                    rawImage[i * width + j] += values[xCoord * width + yCoord];
                    if (newImageValues[i * width + j] > 255) {
                        newImageValues[i * width + j] = 255;
                    }
                }
            }
        }
    }

    private double[] calculateStackedImage() {
        // calculate average image center
        int x = 0;
        int y = 0;
        for (int i = 0; i < imagesInRotation; i++) {
            x += getImageFromRotation(i).centerX;
            y += getImageFromRotation(i).centerY;
        }
        x = x / imagesInRotation;
        y = y / imagesInRotation;
        // calculate stacked image
        Arrays.fill(newImageValues,0);
        Arrays.fill(rawImageValues,0);
        for (int i = 0; i < imagesInRotation; i++) {
            getImageFromRotation(i).stackImages(x,y,newImageValues,rawImageValues);
        }
        return newImageValues;
    }

    private ImageDataWrapper getImageFromRotation(int index) {
        return images[(lastImageIndex + index) % imagesInRotation];
    }

    private void addImageToRotation(double[] values, int width, int height) {
        if (imagesInRotation < numOfImages) {
            imagesInRotation++;
        }
        images[lastImageIndex % imagesInRotation] = new ImageDataWrapper(values,width,height);
        lastImageIndex++;
    }

    /**
     * Set number of images used for stacking calculation.
     *
     * @param newNumber of images used
     */
    public void setNumOfImages(int newNumber) {
        if (newNumber < 1) {
            newNumber = 1;
        }
        numOfImages = newNumber;
        resetImagesRotation();
    }

    private void resetImagesRotation() {
        images = new ImageDataWrapper[numOfImages];
        imagesInRotation = 0;
        lastImageIndex = 0;
    }

    /**
     * Get number of images used for stacking calculation.
     *
     * @return number of images used
     */
    public int getNumOfImages() {
        return numOfImages;
    }

    @Deprecated
    public InterferometerImageAnalyzer(boolean verticalInterferometer) {
        this(verticalInterferometer ? Type.VERTICAL : Type.HORIZONTAL);
    }

    /**
     * Construct new analyzer. The type parameter specifies which sides should be fitted with a function and which
     * function to use (trigonometric or bessel).
     *
     * @param type the type of analysis to peform on each image
     */
    public InterferometerImageAnalyzer(Type type) {
        this.interferometerType = type;
        this.useBesselFunction = interferometerType == Type.BESSEL_HORIZONTAL
                || interferometerType == Type.BESSEL_VERTICAL || interferometerType == Type.BESSEL;
        setNumOfImages(5);
    }

    @Override
    public void clear() {
        previousStartValuesIf = null;
        previousStartValuesIf2 = null;
    }

    @Override
    protected AnalysisImage doAnalysis(IMAGE newImage, double[] values, int iW, int iH, int rX, int rY, int rW, int rH,
            int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold, boolean calculateThreshold,
            boolean performFit, boolean smoothing) {
        double[] newValues;
        IMAGE correctedImage;
        if(numOfImages == 1) {
            newValues = values;
            correctedImage = newImage;
        } else {
            if(oldImageWidth != iW || oldImageHeight != iH) {
                newImageValues = new double[values.length];
                rawImageValues = new double[values.length];
                oldImageWidth = iW;
                oldImageHeight = iH;
                newImageData = new int[values.length];
                newBufferedImage = TBufferedImage.toBufferedImage(newImage);
                resetImagesRotation();
            }

            // calculate average image center
            addImageToRotation(values, iW, iH);
            newValues = calculateStackedImage();

            // Calculate stacked image
            encode(newImageData,newValues);
            newBufferedImage.setRGB(0,0,iW,iH,newImageData,0,iW);
            correctedImage = TBufferedImage.toIMAGE(newBufferedImage);
            correctedImage.getFrameHeader().frameNumber = newImage.getFrameHeader().frameNumber;
            // we need to use the raw image values, which were not cut off at 255
            newValues = rawImageValues;
        }

        AnalysisImage image = null;
        if (interferometerType == Type.VERTICAL || interferometerType == Type.BESSEL_VERTICAL) {
            image = verticalAnalysis(correctedImage,newValues,iW,iH,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,thresholdPointsCount,threshold,
                    calculateThreshold,performFit,smoothing);
        } else if (interferometerType == Type.HORIZONTAL || interferometerType == Type.BESSEL_HORIZONTAL) {
            image = horizontalAnalysis(correctedImage,newValues,iW,iH,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,thresholdPointsCount,threshold,
                    calculateThreshold,performFit,smoothing);
        } else if (interferometerType == Type.BOTH || interferometerType == Type.BESSEL) {
            image = doubleSidedAnalysis(correctedImage,newValues,iW,iH,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,thresholdPointsCount,
                    threshold,calculateThreshold,performFit,smoothing);
        }
        return image;
    }
    
    /**
     * Encode the data into the ret array by creating a grey RGB as specified by the data element.
     *
     * @param ret the destination array
     * @param data the source array
     * @return ret
     */
    protected static int[] encode(int[] ret, double[] data) {
        int length = data.length;
        int code;
        int alpha = (255 << 24) & 0xFF000000;
        for (int i = 0; i < length; i++) {
            code = ((int)(data[i] < 0 ? 0 : data[i] + 0.5)) & 0xFF;
            if (code < 0) code = 0;
            ret[i] = alpha | (code << 16) | (code << 8) | (code);
        }
        return ret;
    }

    private double[] tryFit(double[] sideView, double[] previousStartValues, boolean bessel) throws Exception {
        int l = sideView.length;
        double[] steps = new double[l];
        double[] weights = new double[l];
        for (int i = 0; i < l; i++) {
            steps[i] = i;
            weights[i] = 1;
        }
        double[] cost = new double[6];
        double[] startValues;
        final MultivariateRealFunction costFunction;
        if (bessel) {
            costFunction = new InterferometerCostFunctionBessel(steps,sideView);
        } else {
            costFunction = new InterferometerCostFunctionTrigonometric(steps,sideView);
        }
        //first try to use the previous values as a strating point
        if (previousStartValues != null) {
            try {
                startValues = FittingUtilities.lmInterferometer(false,steps,sideView,weights,previousStartValues,
                        bessel);
                cost[0] = Math.sqrt(costFunction.value(startValues) / l);
                if (cost[0] < COST_MIN) {
                    currentCostMin = COST_MIN;
                }
                if (cost[0] < currentCostMin && !isResultInvalid(startValues)) {
                    if (hasSmallValue(startValues)) return null;
                    return startValues;
                }
            } catch (Exception e) {
                //ignore
            }
            try {
                //if levenberg fails, try nelder mead
                startValues = FittingUtilities.lmInterferometer(true,steps,sideView,weights,previousStartValues,bessel);
                cost[1] = Math.sqrt(costFunction.value(startValues) / l);
                if (cost[1] < COST_MIN) {
                    currentCostMin = COST_MIN;
                }
                if (cost[1] < currentCostMin && !isResultInvalid(startValues)) {
                    if (hasSmallValue(startValues)) return null;
                    return startValues;
                }
            } catch (Exception e) {
                //ignore
            }
            //a fix for when the cost function value increased, just because of a longer exposure time on the camera.
            if (Math.abs(cost[1] - cost[0]) < DELTA) {
                currentCostMin += COST_MIN;
            }
        }
        //estimates start values and try levenberg
        double[] freshStart = getStartValues(sideView,false);
        try {
            startValues = FittingUtilities.lmInterferometer(false,steps,sideView,weights,freshStart,bessel);
            cost[2] = Math.sqrt(costFunction.value(startValues) / l);
            if (cost[2] < currentCostMin && !isResultInvalid(startValues)) {
                if (hasSmallValue(startValues)) return null;
                return startValues;
            }
        } catch (Exception e) {
            //ignore
        }
        //estimates start values and try levenberg
        startValues = getStartValues(sideView,true);
        try {
            startValues = FittingUtilities.lmInterferometer(false,steps,sideView,weights,freshStart,bessel);
            cost[3] = Math.sqrt(costFunction.value(startValues) / l);
            if (Math.abs(cost[3] - cost[2]) < DELTA) {
                currentCostMin += COST_MIN;
            }
            if (cost[2] < currentCostMin && !isResultInvalid(startValues)) {
                if (hasSmallValue(startValues)) return null;
                return startValues;
            }
        } catch (Exception e) {
            //ignore
        }
        //try nelder mead with estimated start values
        startValues = freshStart;
        try {
            startValues = FittingUtilities.lmInterferometer(true,steps,sideView,weights,freshStart,bessel);
            cost[3] = Math.sqrt(costFunction.value(startValues) / l);
            if (cost[3] < currentCostMin && !isResultInvalid(startValues)) {
                if (hasSmallValue(startValues)) return null;
                return startValues;
            }
        } catch (Exception e) {
            //ignore
        }
        startValues = freshStart;
        try {
            startValues = FittingUtilities.lmInterferometer(true,steps,sideView,weights,freshStart,bessel);
            cost[4] = Math.sqrt(costFunction.value(startValues) / l);
            if (Math.abs(cost[4] - cost[3]) < DELTA) {
                currentCostMin += COST_MIN;
            }
        } catch (Exception e) {
            //ignore
        }
        //randomize start values and try levenberg
        freshStart = getStartValues(sideView,true);
        try {
            startValues = FittingUtilities.lmInterferometer(false,steps,sideView,weights,freshStart,bessel);
            cost[4] = Math.sqrt(costFunction.value(startValues) / l);
            if (cost[4] < currentCostMin && !isResultInvalid(startValues)) {
                if (hasSmallValue(startValues)) return null;
                return startValues;
            }
        } catch (Exception e) {
            //ignore
        }
        //randomize start values and try nelder mead
        freshStart = getStartValues(sideView,true);
        try {
            startValues = FittingUtilities.lmInterferometer(true,steps,sideView,weights,freshStart,bessel);
            cost[5] = Math.sqrt(costFunction.value(startValues) / l);
            if (cost[5] < currentCostMin && !isResultInvalid(startValues)) {
                if (hasSmallValue(startValues)) return null;
                return startValues;
            }
        } catch (Exception e) {
            //ignore
        }
        if (Math.abs(cost[5] - cost[4]) < DELTA) {
            currentCostMin += COST_MIN;
        }
        //if nothing converged return null
        return null;
    }

    private AnalysisImage verticalAnalysis(IMAGE newImage, double[] values, int iW, int iH, int rX, int rY, int rW,
            int rH, int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold,
            boolean calculateThreshold, boolean performFit, boolean smoothing) {
        int i, j;
        ImageDataWrapper imageWrapper = new ImageDataWrapper(values, iW, iH);
        double[] sideViewX = new double[iW];
        double[] sideViewY = new double[iH];
        double value;
        if (Double.isNaN(threshold)) {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    sideViewX[i] += values[i + j * iW];
                }
            }
        } else {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (value < threshold) continue;
                    sideViewX[i] += (value - threshold);
                }
            }
        }
        //smoothing
        if (smoothing) {
            sideViewX = FittingUtilities.smooth(sideViewX);
        }
        //end of smoothing
        try {
            i = 0;
            for (;;) {
                if (values[i] < 0) values[i] = 0;
                i++;
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            //ignore
        }
//        int xindex = (int)findMinMax(sideViewX,0,iW)[2];
        int xindex = imageWrapper.centerX;
        rX = xindex - HALFROIWIDTH;
        if (rX < 0) rX = 1;
        rW = 2 * HALFROIWIDTH + 1;
        if (rX + rW > iW) {
            int d = Math.min(HALFROIWIDTH,iW - 1 - xindex);
            if (d < 0) {
                rX = xindex;
                rW = 0;
            } else {
                rX = xindex - d;
                rW = 2 * d + 1;
            }
        }
        double[] roiSideViewY = new double[rH];
        if (Double.isNaN(threshold)) {
            for (i = rX; i < rX + rW; ++i) {
                for (j = 0; j < iH; ++j) {
                    sideViewY[j] += values[i + j * iW];
                }
            }
        } else {
            for (i = rX; i < rX + rW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (value < threshold) continue;
                    sideViewY[j] += (value - threshold);
                }
            }
        }
        try {
            i = 0;
            for (;;) {
                sideViewY[i] = sideViewY[i] / rW;
                i++;
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            //ignore
        }
        if (smoothing) {
            sideViewY = FittingUtilities.smooth(sideViewY);
        }
        System.arraycopy(sideViewY,rY,roiSideViewY,0,rH);
        double[] startValuesY;
        if (performFit) {
            try {
                startValuesY = tryFit(roiSideViewY,previousStartValuesIf2,useBesselFunction);
                if (startValuesY != null) {
                    previousStartValuesIf2 = startValuesY;
                }
            } catch (Exception e) {
                startValuesY = NAN_ARRAY;
                previousStartValuesIf2 = null;
            }
            if (FittingUtilities.isNaN(startValuesY)) {
                startValuesY = NAN_ARRAY;
                previousStartValuesIf2 = null;
            }
        } else {
            startValuesY = NAN_ARRAY;
            previousStartValuesIf2 = null;
        }
        if (useBesselFunction) {
            return new SingleBesselIFMImage(newImage,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,sideViewX,sideViewY,
                    startValuesY[0],startValuesY[1],startValuesY[2],startValuesY[3],startValuesY[4],startValuesY[5],
                    startValuesY[6],thresholdPointsCount,calculateThreshold,performFit,smoothing,true);
        } else {
            return new IFMImage(newImage,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,sideViewX,sideViewY,startValuesY[0],
                    startValuesY[1],startValuesY[2],startValuesY[3],startValuesY[4],startValuesY[5],startValuesY[6],
                    thresholdPointsCount,calculateThreshold,performFit,smoothing,true);
        }
    }

    private AnalysisImage horizontalAnalysis(IMAGE newImage, double[] values, int iW, int iH, int rX, int rY, int rW,
            int rH, int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold,
            boolean calculateThreshold, boolean performFit, boolean smoothing) {
        int i, j;
        ImageDataWrapper imageDataWrapper = new ImageDataWrapper(values, iW, iH);
        double[] sideViewX = new double[iW];
        double[] sideViewY = new double[iH];
        double value;
        if (Double.isNaN(threshold)) {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    sideViewY[j] += values[i + j * iW];
                }
            }
        } else {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (value < threshold) continue;
                    sideViewY[j] += (value - threshold);
                }
            }
        }
        //smoothing
        if (smoothing) {
            sideViewY = FittingUtilities.smooth(sideViewY);
        }
        //end of smoothing
        try {
            i = 0;
            for (;;) {
                if (values[i] < 0) values[i] = 0;
                i++;
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            //ignore
        }
//        int yindex = (int)findMinMax(sideViewY,0,iH)[2];
        int yindex = imageDataWrapper.centerY;
        rY = yindex - HALFROIWIDTH;
        if (rY < 0) rY = 1;
        rH = 2 * HALFROIWIDTH + 1;
        if (rY + rH > iH) {
            int d = Math.min(HALFROIWIDTH,iH - 1 - yindex);
            if (d < 0) {
                rY = yindex;
                rH = 0;
            } else {
                rY = yindex - d;
                rH = 2 * d + 1;
            }
        }
        double[] roiSideViewX = new double[rW];
        if (Double.isNaN(threshold)) {
            for (i = rY; i < rY + rH; ++i) {
                for (j = 0; j < iW; ++j) {
                    sideViewX[j] += values[i * iW + j];
                }
            }
        } else {
            for (i = rY; i < rY + rH; ++i) {
                for (j = 0; j < iW; ++j) {
                    value = values[i * iW + j];
                    if (value < threshold) continue;
                    sideViewX[j] += (value - threshold);
                }
            }
        }
        try {
            i = 0;
            for (;;) {
                sideViewX[i] = sideViewX[i] / rH;
                i++;
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            //ignore
        }
        if (smoothing) {
            sideViewX = FittingUtilities.smooth(sideViewX);
        }
        System.arraycopy(sideViewX,rX,roiSideViewX,0,rW);
        double[] startValuesX;
        if (performFit) {
            try {
                startValuesX = tryFit(roiSideViewX,previousStartValuesIf,useBesselFunction);
                if (startValuesX != null) {
                    previousStartValuesIf = startValuesX;
                }
            } catch (Exception e) {
                startValuesX = NAN_ARRAY;
                previousStartValuesIf = null;
            }
            if (FittingUtilities.isNaN(startValuesX)) {
                startValuesX = NAN_ARRAY;
                previousStartValuesIf = null;
            }
        } else {
            startValuesX = NAN_ARRAY;
            previousStartValuesIf = null;
        }
        if (useBesselFunction) {
            return new SingleBesselIFMImage(newImage,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,sideViewX,sideViewY,
                    startValuesX[0],startValuesX[1],startValuesX[2],startValuesX[3],startValuesX[4],startValuesX[5],
                    startValuesX[6],thresholdPointsCount,calculateThreshold,performFit,smoothing,false);
        } else {
            return new IFMImage(newImage,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,sideViewX,sideViewY,startValuesX[0],
                    startValuesX[1],startValuesX[2],startValuesX[3],startValuesX[4],startValuesX[5],startValuesX[6],
                    thresholdPointsCount,calculateThreshold,performFit,smoothing,false);
        }
    }

    private AnalysisImage doubleSidedAnalysis(IMAGE newImage, double[] values, int iW, int iH, int orgRX, int orgRY,
            int orgRW, int orgRH, int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold,
            boolean calculateThreshold, boolean performFit, boolean smoothing) {
        int i, j;
        double value;
        double[] sideViewX = new double[iW];
        double[] sideViewY = new double[iH];
        ImageDataWrapper imageWrapper = new ImageDataWrapper(values, iW, iH);

//        //put together the side view, to find the min/max
//        for (i = 0; i < iW; ++i) {
//            for (j = 0; j < iH; ++j) {
//                value = values[i + j * iW];
//                sideViewY[j] += (value);
//                sideViewX[i] += (value);
//            }
//        }
//        // Find the the top of the interferometer and get the 10 pixel strip in X direction
//        int yindex = (int)findMinMax(sideViewY,0,iH)[2];
        int yindex = imageWrapper.centerY;
        int rW = orgRW;
        int rH = orgRH;
        int rX = orgRX;
        int rY = orgRY;
        double[] roiSideViewX = new double[rW];
        double[] roiSideViewY = new double[rH];
        rY = yindex - HALFROIWIDTH;
        if (rY < 0) rY = 1;
        rH = 2 * HALFROIWIDTH + 1;
        if (rY + rH > iH) {
            int d = Math.min(HALFROIWIDTH,iH - 1 - yindex);
            if (d < 0) {
                rY = yindex;
                rH = 0;
            } else {
                rY = yindex - d;
                rH = 2 * d + 1;
            }
        }
        // Find the the top of the interferometer and get the 10 pixel strip in Y direction
//        int xindex = (int)findMinMax(sideViewX,0,iW)[2];
        int xindex = imageWrapper.centerX;
        rX = xindex - HALFROIWIDTH;
        if (rX < 0) rX = 1;
        rW = 2 * HALFROIWIDTH + 1;
        if (rX + rW > iW) {
            int d = Math.min(HALFROIWIDTH,iW - 1 - xindex);
            if (d < 0) {
                rX = xindex;
                rW = 0;
            } else {
                rX = xindex - d;
                rW = 2 * d + 1;
            }
        }
        // we need an empty side view to count
        Arrays.fill(sideViewX,0);
        Arrays.fill(sideViewY,0);
        if (Double.isNaN(threshold)) {
            for (i = rY; i < rY + rH; ++i) {
                for (j = 0; j < iW; ++j) {
                    sideViewX[j] += values[i * iW + j];
                }
            }
        } else {
            for (i = rY; i < rY + rH; ++i) {
                for (j = 0; j < iW; ++j) {
                    value = values[i * iW + j];
                    if (value < threshold) continue;
                    sideViewX[j] += (value - threshold);
                }
            }
        }
        try {
            i = 0;
            for (;;) {
                sideViewX[i] = sideViewX[i] / rH;
                i++;
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            //ignore
        }
        if (Double.isNaN(threshold)) {
            for (i = rX; i < rX + rW; ++i) {
                for (j = 0; j < iH; ++j) {
                    sideViewY[j] += values[i + j * iW];
                }
            }
        } else {
            for (i = rX; i < rX + rW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (value < threshold) continue;
                    sideViewY[j] += (value - threshold);
                }
            }
        }
        try {
            i = 0;
            for (;;) {
                sideViewY[i] = sideViewY[i] / rW;
                i++;
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            //ignore
        }
        // Smooth and fit.
        if (smoothing) {
            sideViewX = FittingUtilities.smooth(sideViewX);
            sideViewY = FittingUtilities.smooth(sideViewY);
        }
        System.arraycopy(sideViewX,orgRX,roiSideViewX,0,orgRW);
        System.arraycopy(sideViewY,orgRY,roiSideViewY,0,orgRH);
        double[] startValuesX, startValuesY;
        if (performFit) {
            try {
                startValuesX = tryFit(roiSideViewX,previousStartValuesIf,useBesselFunction);
                if (startValuesX != null) {
                    previousStartValuesIf = startValuesX;
                }
            } catch (Exception e) {
                startValuesX = NAN_ARRAY;
                previousStartValuesIf = null;
            }
            if (FittingUtilities.isNaN(startValuesX)) {
                startValuesX = NAN_ARRAY;
                previousStartValuesIf = null;
            }
            try {
                startValuesY = tryFit(roiSideViewY,previousStartValuesIf2,useBesselFunction);
                if (startValuesY != null) {
                    previousStartValuesIf2 = startValuesY;
                }
            } catch (Exception e) {
                startValuesY = NAN_ARRAY;
                previousStartValuesIf2 = null;
            }
            if (FittingUtilities.isNaN(startValuesY)) {
                startValuesY = NAN_ARRAY;
                previousStartValuesIf2 = null;
            }
        } else {
            startValuesX = NAN_ARRAY;
            previousStartValuesIf = null;
            startValuesY = NAN_ARRAY;
            previousStartValuesIf2 = null;
        }
        if (useBesselFunction) {
            return new BesselIFMImage(newImage,orgRX,orgRY,orgRW,orgRH,r2X,r2Y,r2W,r2H,threshold,sideViewX,sideViewY,
                    startValuesY[0],startValuesY[1],startValuesY[2],startValuesY[3],startValuesY[4],startValuesY[5],
                    startValuesY[6],startValuesX[0],startValuesX[1],startValuesX[2],startValuesX[3],startValuesX[4],
                    startValuesX[5],startValuesX[6],thresholdPointsCount,calculateThreshold,performFit,smoothing);
        } else {
            return new DIFMImage(newImage,orgRX,orgRY,orgRW,orgRH,r2X,r2Y,r2W,r2H,threshold,sideViewX,sideViewY,
                    startValuesY[0],startValuesY[1],startValuesY[2],startValuesY[3],startValuesY[4],startValuesY[5],
                    startValuesY[6],startValuesX[0],startValuesX[1],startValuesX[2],startValuesX[3],startValuesX[4],
                    startValuesX[5],startValuesX[6],thresholdPointsCount,calculateThreshold,performFit,smoothing);
        }
    }

    /**
     * Checks if the result is invalid. Result is invalid if the visibility is less than 0 or if any of the elements is
     * too large.
     *
     * @param result the result to check
     * @return true if the result is invalid or false if valid
     */
    private static boolean isResultInvalid(double[] result) {
        if (result[4] < 0) {
            result[4] = -result[4];
            result[6] = result[6] + Math.PI / result[5];
        }
        for (int i = 0; i < result.length; i++) {
            if (Math.abs(result[i]) > 1E10) return true;
        }
        return false;
    }

    private static boolean hasSmallValue(double[] result) {
        for (int i = 0; i < result.length; i++) {
            if (Math.abs(result[i]) < 1E-30) return true;
        }
        return false;
    }

    /**
     * Estimates the starting values for the fit.
     *
     * @param sideView the profile which the fit is calculated for
     * @return the starting values
     */
    private static double[] getStartValues(double[] sideView, boolean randomize) {
        double[] yMinMax = findMinMax(sideView,0,sideView.length);
        double fr = getFrequency(sideView);
        double bcg = Math.max(yMinMax[0],Math.random() * 3);
        double amp = yMinMax[1] - bcg;
        double asin = 0.014;
        double psin = yMinMax[2];
        double vis = 0.9;
        double acos = fr;//(fr > 0.1 && fr < 0.4) ? fr : 0.3;
        double pcos = 200;
        if (randomize) {
            vis = 0.2 + Math.random() * 0.75;
            asin = Math.random() * 0.02;
            pcos = 100 + Math.random();
        }
        return new double[]{bcg,amp,asin,psin,vis,acos,pcos};
    }

    /**
     * Calculates the most important frequency that is found in the given function.
     *
     * @param y the function
     * @return the frequency
     */
    private static double getFrequency(double[] y) {
        double[] f = null;
        int l = y.length - 1;
        //get power of 2
        l |= l >> 1;
        l |= l >> 2;
        l |= l >> 4;
        l |= l >> 8;
        l |= l >> 16;
        l++;
        f = new double[l];
        System.arraycopy(y,0,f,0,y.length);
        double[][] fft = FittingUtilities.FFT1D(new double[][]{f,new double[l]},true);
        int index = 0;
        double max = 0;
        double t;
        double tm1 = Math.sqrt(fft[0][0] * fft[0][0] + fft[1][0] * fft[1][0]);
        double tp1 = Math.sqrt(fft[0][2] * fft[0][2] + fft[1][2] * fft[1][2]);
        t = Math.sqrt(fft[0][1] * fft[0][1] + fft[1][1] * fft[1][1]);
        for (int i = 10; i < l / 2; i++) {
            if (t > max && t > tm1 && t > tp1) {
                max = t;
                index = i - 1;
            }
            tm1 = t;
            t = tp1;
            tp1 = Math.sqrt(fft[0][i] * fft[0][i] + fft[1][i] * fft[1][i]);
        }
        return index * 2 * Math.PI / l;
    }

    @Override
    public AnalysisType getAnalysisType() {
        switch (interferometerType) {
            case BESSEL:
                return AnalysisType.INTERFEROMETER_BESSEL;
            case BESSEL_HORIZONTAL:
                return AnalysisType.INTERFEROMETER_BESSEL_H;
            case BESSEL_VERTICAL:
                return AnalysisType.INTERFEROMETER_BESSEL_V;
            case BOTH:
                return AnalysisType.INTERFEROMETER_BOTH;
            case HORIZONTAL:
                return AnalysisType.INTERFEROMETER_H;
            case VERTICAL:
                return AnalysisType.INTERFEROMETER_V;
            default:
                return AnalysisType.INTERFEROMETER_BOTH;
        }
    }
}
